# -*- coding: utf-8 -*-

"""Console script for BitcoinExchangeFH."""

import logging

import click
import yaml

from befh import Configuration, Runner

LOGGER = logging.getLogger(__name__)


@click.command()
@click.option(
    '--configuration',
    help='Configuration file.',
    required=True)
@click.option(
    '--debug',
    default=False,
    is_flag=True,
    help='Debug mode.')
@click.option(
    '--cold',
    default=False,
    is_flag=True,
    help='Cold start mode.')
@click.option(
    '--archive',
    default=None,
    help='Manually archive the tables.',
    required=False)
def main(configuration, debug, cold, archive):
    """Console script for BitcoinExchangeFH."""
    if debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(
        level=level,
        format='%(asctime)s %(levelname)s %(message)s')

    configuration = open(configuration, 'r')
    configuration = yaml.safe_load(configuration, Loader=yaml.FullLoader)
    LOGGER.debug('Configuration:\n%s', configuration)
    configuration = Configuration(configuration)

    runner = Runner(
        config=configuration,
        is_debug=debug,
        is_cold=cold)
    runner.load()

    if archive is not None:
        runner.archive(date=archive)
    else:
        runner.run()


if __name__ == "__main__":
    main()



import logging
from datetime import datetime
from time import sleep

import ccxt
from ccxt.base.errors import RequestTimeout, NetworkError, ExchangeError

from .exchange import Exchange

LOGGER = logging.getLogger(__name__)


class RestApiExchange(Exchange):
    """Rest API exchange.
    """

    def load(self, is_initialize_instmt=True, **kwargs):
        """Load.
        """
        super().load(**kwargs)
        ccxt_exchange = getattr(ccxt, self._name.lower(), None)
        if ccxt_exchange:          
            self._exchange_interface = ccxt_exchange()
            self._exchange_interface.load_markets()
            self._check_valid_instrument()
            if is_initialize_instmt:
                self._initialize_instmt_info()

    def run(self):
        """Run.
        """
        while True:
            for symbol, instmt_info in self._instruments.items():
                self._update_order_book(
                    symbol=symbol,
                    instmt_info=instmt_info)

                if instmt_info.is_possible_trade():
                    self._update_trades(
                        symbol=symbol,
                        instmt_info=instmt_info)

                self._rotate_ordre_tables()

    def _check_valid_instrument(self):
        """Check valid instrument.
        """
        for instrument_code in self._config['instruments']:
            if instrument_code not in self._exchange_interface.markets:
                raise RuntimeError(
                    'Instrument %s is not found in exchange %s',
                    instrument_code, self._name)

    def _initialize_instmt_info(self):
        """Initialize instrument info.
        """
        for symbol, instmt_info in self._instruments.items():
            self._update_order_book(
                symbol=symbol,
                instmt_info=instmt_info,
                is_update_handler=False)

            self._update_trades(
                symbol=symbol,
                instmt_info=instmt_info,
                is_update_handler=False)

    def _update_order_book(self, symbol, instmt_info, is_update_handler=True):
        """Callback order book.
        """
        tolerence_count = 0
        order_book = None

        while tolerence_count < self.TIMEOUT_TOLERANCE:
            self._load_balance()
            try:
                order_book = self._exchange_interface.fetch_order_book(
                    symbol=symbol)
                break
            except (RequestTimeout, NetworkError, ExchangeError) as e:
                tolerence_count += 1
                LOGGER.warning('Request timeout %s', e)

        if order_book is None:
            raise RuntimeError(
                'Cannot load the order book after failover. '
                'Please check the exceptions before and network connection')

        bids = order_book['bids']
        asks = order_book['asks']

        is_updated = instmt_info.update_bids_asks(
            bids=bids,
            asks=asks)

        if not is_updated:
            return

        if is_update_handler:
            for handler in self._handlers.values():
                instmt_info.update_table(handler=handler)

    def _update_trades(self, symbol, instmt_info, is_update_handler=True):
        """Update trades.
        """
        tolerence_count = 0
        trades = None

        while tolerence_count < self.TIMEOUT_TOLERANCE:
            self._load_balance()
            try:
                trades = self._exchange_interface.fetch_trades(symbol=symbol)
                break
            except (RequestTimeout, NetworkError, ExchangeError) as e:
                tolerence_count += 1
                LOGGER.warning('Request timeout %s', e)

        if trades is None:
            raise RuntimeError(
                'Cannot load the trades after failover. '
                'Please check the exceptions before and network connection')

        current_timestamp = datetime.utcnow()

        for trade in trades:
            if not instmt_info.update_trade(trade, current_timestamp):
                continue

            if is_update_handler:
                for handler in self._handlers.values():
                    instmt_info.update_table(handler=handler)

    def _rotate_ordre_tables(self):
        """Rotate order table.
        """
        for name, handler in self._handlers.items():
            if not handler.is_rotate:
                continue

            current_timestamp = datetime.utcnow()
            if handler.should_rotate(current_timestamp):
                for instmt_info in self._instruments.values():
                    handler.rotate_table(
                        table=instmt_info,
                        last_datetime=handler.last_rotated_timestamp)

                handler.update_last_rotate_timestamp(current_timestamp)

    def _load_balance(self):
        """Load balance.
        """
        current_time = datetime.now()
        time_diff = (current_time -
                     self._last_request_time).microseconds / 1000.0

        # Rate limit is represented as microseconds,
        # number of requests per seconds = 1000 / rateLimit
        if time_diff < self._exchange_interface.rateLimit:
            wait_second = (
                self._exchange_interface.rateLimit - time_diff) / 1000.0
            sleep(wait_second)

        self._last_request_time = datetime.now()



import logging
from datetime import datetime

from befh.table.order_book_table import OrderBook

LOGGER = logging.getLogger(__name__)


class Exchange:
    """Exchange.
    """

    DEFAULT_ORDER_BOOK_CLASS = OrderBook
    TIMEOUT_TOLERANCE = 5
    DEFAULT_DEPTH = 5
    DEFAULT_TYPE = 'spot'

    def __init__(self, name, config, is_debug, is_cold):
        """Constructor.
        """
        self._name = name
        self._config = config
        self._is_debug = is_debug
        self._is_cold = is_cold
        self._is_orders = True
        self._instruments = {}
        self._depth = Exchange.DEFAULT_DEPTH
        self._type = Exchange.DEFAULT_TYPE
        self._last_request_time = datetime(1990, 1, 1)
        self._exchange_interface = None
        self._handlers = {}

    @classmethod
    def get_order_book_class(cls):
        """Get order book class.
        """
        return OrderBook

    @property
    def name(self):
        """Name.
        """
        return self._name

    @property
    def instruments(self):
        """Instruments.
        """
        return self._instruments

    @property
    def handlers(self):
        """Handlers.
        """
        return self._handlers

    def load(self, handlers, **kwargs):
        """Load.
        """
        LOGGER.info('Loading exchange %s', self._name)
        self._load_handlers(handlers=handlers)
        self._load_instruments()
        self._load_depth()
        self._load_type()
        self._load_is_orders()

    def _load_handlers(self, handlers):
        """Load handlers.
        """
        self._handlers = handlers

    def _load_instruments(self):
        """Load instruments.
        """
        instruments = self._config['instruments']
        for symbol in instruments:
            instmt_info = self.DEFAULT_ORDER_BOOK_CLASS(
                exchange=self._name,
                symbol=self._symbol_filter(symbol))
            self._instruments[symbol] = instmt_info

            for handler in self._handlers.values():
                handler.prepare_create_table(
                    table_name=instmt_info.table_name,
                    fields=instmt_info.fields)

    def _load_depth(self):
        """Load depth.
        """
        if 'depth' in self._config:
            self._depth = self._config['depth']
            assert isinstance(self._depth, int), (
                "Depth ({}) must be an integer".format(
                    self._depth))
            
    def _load_type(self):
        """Load type.
        """
        if 'type' in self._config:
            self._type = self._config['type']
            assert isinstance(self._type, str), (
                "Type ({}) must be an string".format(
                    self._type))    
        
    def _load_is_orders(self):
        """Load is_orders.
        """
        if 'is_orders' in self._config:
            self._is_orders = self._config['is_orders']
            assert isinstance(self._is_orders, bool), (
                "is_orders ({}) must be an boolean".format(
                    self._is_orders))    

    @staticmethod
    def _symbol_filter(original_symbol):     
        if original_symbol.find(':')>=0:
            names = original_symbol.split(':')
            symbol = names[1]
        else:
            symbol = original_symbol
            
        return symbol
        


from datetime import datetime
import logging

from .handler import Handler

LOGGER = logging.getLogger(__name__)


class RotateHandler(Handler):
    """Rotate handler.
    """

    def __init__(self, is_rotate=False,
                 rotate_frequency='%Y%m%d',
                 **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._is_rotate = is_rotate
        self._rotate_frequency = rotate_frequency
        self._last_rotated_timestamp = None

    @property
    def is_rotate(self):
        """Is rotate.
        """
        return self._is_rotate

    @property
    def rotate_frequency(self):
        """Rotate frequency.
        """
        return self._rotate_frequency

    @property
    def last_rotated_timestamp(self):
        """Last rotated timestamp.
        """
        return self._last_rotated_timestamp

    def load(self, **kwargs):
        """Load.
        """
        super().load(**kwargs)
        self._last_rotated_timestamp = datetime.utcnow()

    def should_rotate(self, timestamp):
        """Check if all the tables should be rotated.
        """
        return (
            timestamp.strftime(self.rotate_frequency) !=
            self._last_rotated_timestamp.strftime(
                self.rotate_frequency))

    def update_last_rotate_timestamp(self, timestamp):
        """Update last rotate timestamp.
        """
        self._last_rotated_timestamp = timestamp

    def rotate_table(self, table, last_datetime, allow_fail=False):
        """Rotate table.
        """
        from_name = table.table_name
        to_name = "%s_%s" % (
            from_name, self.last_rotated_timestamp.strftime(
                self.rotate_frequency))

        LOGGER.info('Rotate table from %s to %s',
                    from_name,
                    to_name)
        self.prepare_rename_table(
            from_name=from_name,
            to_name=to_name,
            fields=table.fields,
            allow_fail=allow_fail,
            keep_table=True)



import zmq

PORT = 9124


def main():
    """Main.
    """
    context = zmq.Context()
    socket = context.socket(zmq.REP)
    print('Binding port %s' % PORT)
    socket.bind("tcp://*:%s" % PORT)

    while True:
        message = socket.recv()
        print(message)
        socket.send(message)


if __name__ == '__main__':
    main()



class Configuration:
    """Configuration.
    """

    def __init__(self, config):
        """Constructor.

        :param config: A `dict` of configuration.
        """
        self._config = config

    @property
    def subscriptions(self):
        """Subscriptions.
        """
        return self._config['subscriptions']

    @property
    def handlers(self):
        """Handlers.
        """
        return self._config['handlers']

    def keys(self):
        """Keys.
        """
        for key in self._config.keys():
            yield key.lower()

    def check_configuration(self):
        """Check configuration.
        """
        self._check_subscriptions()
        self._check_handlers()

    def _check_subscriptions(self):
        """Check subscriptions.
        """
        if 'subscriptions' not in self.keys():
            raise RuntimeError(
                'No subscription section is found in the configuration')

        if not isinstance(self.subscriptions, dict):
            raise RuntimeError(
                'Subscription handler must be a dict.')

        for exchange, subscription in self.subscriptions.items():
            if not isinstance(subscription, dict):
                raise RuntimeError(
                    'No subscription information is found in '
                    'exchange %s', exchange)

            if 'instruments' not in subscription:
                raise RuntimeError(
                    'Subscription instruments is not found in '
                    'exchange %s. Please state the "instruments" '
                    'for subscription',
                    exchange)

    def _check_handlers(self):
        """Check handlers.
        """
        if 'handlers' not in self.keys():
            raise RuntimeError(
                'No handler section is found in the configuration')

        if not isinstance(self.handlers, dict):
            raise RuntimeError(
                'Handler section must be a dict.')



import logging
from datetime import datetime
import re

from cryptofeed import FeedHandler
from cryptofeed.defines import L2_BOOK, TRADES, BID, ASK
import cryptofeed.exchanges as cryptofeed_exchanges

from .rest_api_exchange import RestApiExchange

LOGGER = logging.getLogger(__name__)


class WebsocketExchange(RestApiExchange):
    """Websocket exchange.
    """

    def __init__(self, **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._feed_handler = None
        self._instrument_mapping = None

    def load(self, **kwargs):
        """Load.
        """
        super().load(is_initialize_instmt=False, **kwargs)
        self._feed_handler = FeedHandler()
        self._instrument_mapping = self._create_instrument_mapping()
        try:
            exchange = getattr(
                cryptofeed_exchanges,
                self._get_exchange_name(self._name))
        except AttributeError as e:
            raise ImportError(
                'Cannot load exchange %s from websocket' % self._name)

        if self._is_orders:                       
            channels = [TRADES, L2_BOOK]            
            callbacks = {
                TRADES: self._update_trade_callback,
                L2_BOOK: self._update_order_book_callback               
            }            
        else:
            channels = [TRADES]
            callbacks = {
                TRADES: self._update_trade_callback,                
            }            

        
        self._feed_handler.add_feed(
            exchange(
                symbols=list(self._instrument_mapping.keys()),
                channels=channels,
                callbacks=callbacks))

    def run(self):
        """Run.
        """
        self._feed_handler.run()

    @staticmethod
    def _get_exchange_name(name):
        """Get exchange name.
        """
        if name == 'Hitbtc':
            return 'HitBTC'
        elif name == 'Okex':
            return "OKEx"
        elif name == "HuobiPro":
            return "Huobi"

        return name

    def _create_instrument_mapping(self):
        """Create instrument mapping.
        """
        mapping = {}
        for name in self._instruments.keys():
            if self._name.lower() == 'bitmex' or self._type == 'futures' or self._type == 'swap':
                # BitMEX uses the instrument name directly
                # without normalizing to cryptofeed convention
                normalized_name = name
            else:
                market = self._exchange_interface.markets[name]
                normalized_name = market['base'] + '-' + market['quote']
                
            mapping[normalized_name] = name

        return mapping

    async def _update_order_book_callback(self, ob, receipt_timestamp):
        """Update order book callback.
        """
        feed = ob.exchange
        pair = ob.symbol
        book = ob.book
        instrument_key = self._get_instrument_key(feed, pair)
            
        instmt_info = self._instruments[instrument_key]
        
        is_updated = instmt_info.websocket_update_bids_asks(
            bids=book.bids,
            asks=book.asks)

        if not is_updated:
            return
        

    async def _update_trade_callback(
            self, t, receipt_timestamp):
        """Update trade callback.
        """
        feed = t.exchange
        pair = t.symbol
        
        instrument_key = self._get_instrument_key(feed, pair)
            
        instmt_info = self._instruments[instrument_key]
        trade = {}

        trade['timestamp'] = t.timestamp
        trade['id'] = t.id
        trade['price'] = float(t.price)
        trade['amount'] = float(t.amount)

        current_timestamp = datetime.utcnow()

        if not instmt_info.update_trade(trade, current_timestamp):
            return

        for handler in self._handlers.values():
            instmt_info.update_table(handler=handler)

        self._rotate_ordre_tables()

    def _check_valid_instrument(self):
        """Check valid instrument.
        """
        skip_checking_exchanges = ['bitmex', 'bitfinex', 'okex']
        if self._name.lower() in skip_checking_exchanges:
            # Skip checking on BitMEX
            # Skip checking on Bitfinex
            return

        for instrument_code in self._config['instruments']:
            if instrument_code not in self._exchange_interface.markets:
                raise RuntimeError(
                    'Instrument %s is not found in exchange %s',
                    instrument_code, self._name)
            
            
    def _get_instrument_key(self, feed, pair):
           
        instrument_key = self._instrument_mapping[pair]  
            
        return instrument_key



import zmq

PORT = 9123


def main():
    """Main.
    """
    context = zmq.Context()
    socket = context.socket(zmq.SUB)
    print('Connecting port %s' % PORT)
    socket.setsockopt(zmq.SUBSCRIBE, b'')
    socket.connect("tcp://localhost:%s" % PORT)
    print('Connected port %s' % PORT)

    while True:
        message = socket.recv()
        print("Message received: %s" % message)


if __name__ == '__main__':
    main()



import logging
from datetime import datetime

from sqlalchemy import (
    create_engine,
    Table,
    Column,
    Integer,
    String,
    Numeric,
    MetaData)

from .rotate_handler import RotateHandler

LOGGER = logging.getLogger(__name__)


class SqlHandler(RotateHandler):
    """Sql handler.
    """

    def __init__(self, connection, **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._connection = connection
        self._engine = None

    @property
    def engine(self):
        """Engine.
        """
        return self._engine

    @property
    def queue(self):
        """Queue.
        """
        return self._queue

    def load(self, queue):
        """Load.
        """
        super().load(queue=queue)
        self._engine = create_engine(self._connection)

    def create_table(self, table_name, fields, **kwargs):
        """Create table.
        """
        assert self._engine, "Engine is not initialized"

        # Check if the table exists
        if table_name in self._engine.table_names():
            if self._is_cold:
                self._engine.execute(
                    'delete table {table_name}'.format(
                        table_name=table_name))
                LOGGER.info(
                    'Table %s is deleted in cold mode',
                    table_name)
            else:
                LOGGER.info('Table %s is created', table_name)
                return

        LOGGER.info('Creating table %s', table_name)
        columns = []

        for field_name, field in fields.items():
            columns.append(self._create_column(
                field_name=field_name,
                field=field))

        meta_data = MetaData()
        Table(table_name, meta_data, *columns)
        meta_data.create_all(self._engine)
        LOGGER.info('Created table %s', table_name)

    def insert(self, table_name, fields):
        """Insert.
        """
        assert self._engine, "Engine is not initialized"

        fields = [
            (k, v) for k, v in fields.items() if not v.is_auto_increment]
        fields = list(zip(*fields))

        column_names = (','.join(fields[0]))
        values = (','.join([str(f) for f in fields[1]]))

        sql_statement = (
            "insert into {table_name} ({column_names}) values "
            "({values})").format(
                table_name=table_name,
                column_names=column_names,
                values=values)

        self._engine.execute(sql_statement)

    def rename_table(self, from_name, to_name, fields=None, keep_table=True):
        """Rename table.
        """
        from alembic.migration import MigrationContext
        from alembic.operations import Operations

        # Refresh the connection again
        self._engine = create_engine(self._connection)
        conn = self._engine.connect()
        ctx = MigrationContext.configure(conn)
        op = Operations(ctx)
        op.rename_table(from_name, to_name)

        if keep_table:
            assert fields is not None, (
                "Fields must be provided to create the table")
            self.create_table(
                table_name=from_name,
                fields=fields)

    @staticmethod
    def _create_column(field_name, field):
        """Create column.
        """
        field_params = {}

        if field.field_type is int:
            field_type = Integer
        elif field.field_type is str:
            field_type = String(field.field_length)
        elif field.field_type is float:
            field_type = Numeric(
                precision=field.size,
                scale=field.decimal)
        elif field.field_type is datetime:
            field_type = String(26)
        else:
            raise NotImplementedError(
                'Field type {type} not implemented'.format(
                    type=field.field_type))

        if field.is_key:
            field_params['primary_key'] = True

        if field.is_auto_increment:
            field_params['autoincrement'] = True

        return Column(field_name, field_type, **field_params)

    def _should_rerun(self, element, exception):
        """Handle exception.
        """
        if element.allow_fail:
            LOGGER.warn(
                'Execution failed on element %s (%s)',
                element,
                str(exception))
            return element.should_rerun
            if element.should_rerun:
                return True
            else:
                return False
        elif 'MySQL server has gone away' in str(exception):
            # Only for MySQL case:
            # Shuold rerun on MySQL server gone
            return True
        else:
            raise



class HandlerOperator:
    """Handler operator.
    """

    def __init__(self, allow_fail=False, should_rerun=False):
        """Constructor.
        """
        self.allow_fail = allow_fail
        self.should_rerun = should_rerun

    def execute(self, handler):
        """Execute.
        """
        raise NotImplementedError(
            'Execute is not implemented')
    
    @staticmethod
    def parse_table_name(table_name):
        """parse table name fix sqlalchemy can't insert table name with '.'.
        """
            
        return table_name.replace('.', '')


class HandlerCloseOperator(HandlerOperator):
    """Handler close operator.
    """

    def execute(self, handler):
        """Execute.
        """
        handler.close()


class HandlerCreateTableOperator(HandlerOperator):
    """Create table operator.
    """

    def __init__(self, table_name, fields, **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._table_name = self.parse_table_name(table_name)
        self._fields = fields

    def execute(self, handler):
        """Execute.
        """
        handler.create_table(
            table_name=self._table_name,
            fields=self._fields)


class HandlerInsertOperator(HandlerOperator):
    """Insert operator.
    """

    def __init__(self, table_name, fields, **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._table_name = self.parse_table_name(table_name)
        self._fields = fields

    def execute(self, handler):
        """Execute.
        """
        handler.insert(
            table_name=self._table_name,
            fields=self._fields)


class HandlerRenameTableOperator(HandlerOperator):
    """Rename table operator.
    """

    def __init__(self, from_name, to_name, fields=None,
                 keep_table=True, **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._from_name = self.parse_table_name(from_name)
        self._to_name = self.parse_table_name(to_name)
        self._fields = fields
        self._keep_table = keep_table

    def execute(self, handler):
        """Execute.
        """
        handler.rename_table(
            from_name=self._from_name,
            to_name=self._to_name,
            fields=self._fields,
            keep_table=self._keep_table)



# -*- coding: utf-8 -*-

"""Top-level package for Bitcoin exchange feedhandler."""

__author__ = """Gavin Chan"""
__email__ = 'gavincyi@gmail.com'

from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

# flake8: noqa
from .core import Configuration, Runner



import logging
from datetime import datetime

import zmq

from .handler import Handler

LOGGER = logging.getLogger(__name__)


class ZmqHandler(Handler):
    """Zmq handler.
    """

    def __init__(self, connection, **kwargs):
        """Constructor.
        """
        super().__init__(**kwargs)
        self._connection = connection
        self._context = zmq.Context()
        self._socket = None

    def load(self, queue):
        """Load.
        """
        super().load(queue=queue)
        LOGGER.info('Binding connection %s as a publisher',
                    self._connection)

    def create_table(self, table_name, fields, **kwargs):
        """Create table.
        """
        assert self._socket, "Socket is not initialized"

    def insert(self, table_name, fields):
        """Insert.
        """
        assert self._socket, "Socket is not initialized"

        native_fields = {
            k: self.serialize(v) for k, v in fields.items()
            if not v.is_auto_increment}

        data = {
            "table_name": table_name,
            "data": native_fields
        }

        self._socket.send_json(data)

    @staticmethod
    def serialize(value):
        """Serialize value.
        """
        if isinstance(value.value, datetime):
            return str(value)

        return value.value

    def run(self):
        """Run.
        """
        # The socket has to be initialized here due to pyzmq #1232
        # https://github.com/zeromq/pyzmq/issues/1232
        self._socket = self._context.socket(zmq.PUB)
        self._socket.bind(self._connection)
        super().run()



from datetime import datetime
from collections import OrderedDict
from copy import deepcopy

from .table import (
    Table,
    Field,
    IntIdField,
    StringIdField,
    DateTimeField,
    PriceField,
    QuantityField)


class OrderBookUpdateTypeField(Field):
    """Order book update type field.
    """

    ORDER_BOOK = 1
    TRADE = 2

    @property
    def field_type(self):
        """Field type.
        """
        return int


class OrderBook(Table):
    """Order book.
    """

    TABLE_NAME = '{exchange}_{symbol}_order'
    DEFAULT_NUM_TIMESTAMPS_STORED = 10

    TRADE_PX_INDEX = 0
    TRADE_QTY_INDEX = 1
    TRADE_ID_INDEX = 2
    TRADE_TIMESTAMP_INDEX = 3

    def __init__(self, exchange, symbol, depth=5):
        """Constructor.

        :param depth: `int` of order book depth.
        """
        self._exchange = exchange
        self._symbol = symbol
        self._depth = depth
        self._bids = self.create_depths('b', depth)
        self._asks = self.create_depths('a', depth)
        self._prev_bids = self.create_depths('b', depth)
        self._prev_asks = self.create_depths('a', depth)
        self._trade = self.create_trade()
        self._prev_trade = self.create_trade()
        self._update_type = OrderBookUpdateTypeField(
            name='update_type',
            value=0)
        self._update_time = DateTimeField(
            name='date_time', value=datetime(2000, 1, 1))
        self._prev_update_time = DateTimeField(
            name='date_time', value=datetime(2000, 1, 1))
        #self._trades_per_timestamp = {}

    @staticmethod
    def create_depths(prefix, depth):
        """Create depths.
        """
        depths = []

        for i in range(1, depth + 1):
            depths.append((
                PriceField(name='%s%d' % (prefix, i), value=-1),
                QuantityField(name='%sq%d' % (prefix, i), value=-1)))

        return depths

    @staticmethod
    def create_trade(price=-1, quantity=-1, id='', timestamp=''):
        """Create trade.
        """
        return [
            PriceField(name='t', value=price),
            QuantityField(name='tq', value=quantity),
            StringIdField(name='tid', value=id),
            IntIdField(name='tts', value=0),
        ]

    @property
    def table_name(self):
        return self.TABLE_NAME.format(
            exchange=self._exchange.lower(),
            symbol=self._symbol.replace('/', '').lower())

    @property
    def fields(self):
        """Fields.
        """
        fields = OrderedDict()
        for field in self._get_fields():
            fields[field.name] = field

        return fields

    def _get_fields(self):
        """Get fields.
        """
        fields = [
            IntIdField(name='id'),
            self._update_time,
            self._update_type,
            self._trade[self.TRADE_PX_INDEX],
            self._trade[self.TRADE_QTY_INDEX]
        ]

        for i in range(0, self._depth):
            fields += self._bids[i]
            fields += self._asks[i]

        return fields

    def create_table(self, handler):
        """Create table.
        """
        handler.create_table(
            table_name=self.table_name,
            fields=self.fields)

    def update_table(self, handler):
        """Update table.
        """
        handler.prepare_insert(
            table_name=self.table_name,
            fields=self.fields)

    def is_possible_trade(self):
        """Check if any trade is detected.
        """
        if self._bids[0][0] != self._prev_bids[0][0]:
            return True
        elif self._bids[0][1] != self._prev_bids[0][1]:
            return True

        if self._bids[0][0] != self._prev_bids[0][0]:
            return True
        elif self._bids[0][1] != self._prev_bids[0][1]:
            return True

        return False

    def update_bids_asks(self, bids, asks):
        """Update bids and asks.
        """
        is_update = False

        self._prev_bids = deepcopy(self._bids)
        self._prev_asks = deepcopy(self._asks)

        # Ensure proper order is defined
        if len(bids) > 1 and bids[0][0] < bids[1][0]:
            bids = bids[::-1]

        if len(asks) > 1 and asks[0][0] > asks[1][0]:
            asks = asks[::-1]

        max_bid_depth = min(len(bids), self._depth)
        max_ask_depth = min(len(asks), self._depth)

        for i in range(0, max_bid_depth):
            self._bids[i][0].value = bids[i][0]
            self._bids[i][1].value = bids[i][1]
            is_update |= (
                self._bids[i][0] != self._prev_bids[i][0] or
                self._bids[i][1] != self._prev_bids[i][1]
            )

        for i in range(0, max_ask_depth):
            self._asks[i][0].value = asks[i][0]
            self._asks[i][1].value = asks[i][1]
            is_update |= (
                self._asks[i][0] != self._prev_asks[i][0] or
                self._asks[i][1] != self._prev_asks[i][1]
            )

        self._prev_update_time.value = self._update_time.value
        self._update_time.value = datetime.utcnow()
        self._update_type.value = (
            OrderBookUpdateTypeField.ORDER_BOOK)

        return is_update
    
    
    def websocket_update_bids_asks(self, bids, asks):
        """Update bids and asks.
        """
        is_update = False

        self._prev_bids = deepcopy(self._bids)
        self._prev_asks = deepcopy(self._asks)

        # Ensure proper order is defined
        """
        if len(bids) > 1 and bids.index(0)[0] < bids.index(1)[0]:
            bids_it = reversed(bids)
        else:
            bids_it = iter(bids)

        if len(asks) > 1 and asks.index(0)[0] > asks.index(1)[0]:
            asks_it = reversed(asks)
        else:
            asks_it = iter(asks)
        """
        max_bid_depth = min(len(bids), self._depth)
        max_ask_depth = min(len(asks), self._depth)

        
            
        for i in range(0, max_bid_depth):

            price_f = float(bids.index(i)[0])
            volume_f = float(bids.index(i)[1])
            self._bids[i][0].value = price_f
            self._bids[i][1].value = volume_f
            is_update |= (
                self._bids[i][0] != self._prev_bids[i][0] or
                self._bids[i][1] != self._prev_bids[i][1]
            )
                  

        for i in range(0, max_ask_depth):
            
            price_f = float(asks.index(i)[0])
            volume_f = float(asks.index(i)[1])            
            self._asks[i][0].value = price_f
            self._asks[i][1].value = volume_f
            is_update |= (
                self._asks[i][0] != self._prev_asks[i][0] or
                self._asks[i][1] != self._prev_asks[i][1]
            )
                          
       
        self._prev_update_time.value = self._update_time.value
        self._update_time.value = datetime.utcnow()
        self._update_type.value = (
            OrderBookUpdateTypeField.ORDER_BOOK)
        
        return is_update

    def update_trade(self, trade, current_timestamp):
        """Update trades.
        """
        timestamp = trade['timestamp']
        trade_id = trade['id']

        if self._trade is not None:
            # If the timestamp is before the latest trade timestamp,
            # the trade must be proceeded before
            if timestamp < self._trade[self.TRADE_TIMESTAMP_INDEX]:
                return False

            # If the timestamp is same as the latest trade timestamp,
            # needs to check whether it is proceeded
            if timestamp == self._trade[self.TRADE_TIMESTAMP_INDEX]:
                # The trade equals the latest trade
                if trade_id == self._trade[self.TRADE_ID_INDEX]:
                    return False

                # Check whether the trade was proceeded before at the
                # same timestamp
                #timestamp_trades = self._trades_per_timestamp[timestamp]

                #if trade_id in timestamp_trades:
                    #return False

        self._prev_trade = deepcopy(self._trade)
        self._trade[self.TRADE_PX_INDEX].value = trade['price']
        self._trade[self.TRADE_QTY_INDEX].value = trade['amount']
        self._trade[self.TRADE_ID_INDEX].value = trade['id']
        self._trade[self.TRADE_TIMESTAMP_INDEX].value = trade['timestamp']
        self._prev_update_time.value = self._update_time.value
        self._update_time.value = current_timestamp
        self._update_type.value = OrderBookUpdateTypeField.TRADE
        #self._trades_per_timestamp.setdefault(timestamp, []).append(
            #trade_id)

        return True



from time import sleep

import zmq

PORT = 9123


def main():
    """Main.
    """
    context = zmq.Context()
    socket = context.socket(zmq.PUB)
    print('Binding port %s' % PORT)
    socket.bind("tcp://*:%s" % PORT)

    while True:
        socket.send(b"a abc")
        sleep(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

requirements = [
    'Click>=6.0',
    'PyYAML>=4.2b1',
    'ccxt>=1.17.485',
    'sqlalchemy>=1.1.11',
    'alembic>=0.9.3',
    'pyzmq>=17.1.2',
]

setup_requirements = [
    'setuptools_scm',
    'pytest-runner',
]

test_requirements = [
    'pytest',
]

extra_requirements = {
    ":python_version>='3.5.3'": ["cryptofeed>=1.4.1"],
}


setup(
    name='BitcoinExchangeFH',
    use_scm_version=True,
    description="Cryptocurrency exchange market data feed handler",
    author="Gavin Chan",
    author_email='gavincyi@gmail.com',
    url='https://github.com/gavincyi/BitcoinExchangeFH',
    packages=find_packages(include=['befh']),
    entry_points={
        'console_scripts': [
            'bitcoinexchangefh=befh.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='BitcoinExchangeFH',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    extras_require=extra_requirements,
)



import logging
import multiprocessing as mp
from datetime import datetime

LOGGER = logging.getLogger(__name__)


class Runner:
    """Runner.
    """

    def __init__(self, config, is_debug, is_cold):
        """Constructor.
        """
        self._config = config
        self._is_debug = is_debug
        self._is_cold = is_cold
        self._exchanges = {}
        self._handlers = {}

    def load(self):
        """Load.
        """
        LOGGER.info('Loading runner')

        handlers_configuration = self._config.handlers
        handlers = self.create_handlers(
            handlers_configuration=handlers_configuration,
            is_debug=self._is_debug,
            is_cold=self._is_cold)

        self._handlers = handlers

        exchanges_configuration = self._config.subscriptions
        exchanges = self.create_exchanges(
            exchanges_configuration=exchanges_configuration,
            handlers=handlers,
            is_debug=self._is_debug,
            is_cold=self._is_cold)

        self._exchanges = exchanges

    def run(self):
        """Run.
        """
        LOGGER.info('Start running the feed handler')

        processes = []

        for name, handler in self._handlers.items():
            LOGGER.info('Running handler %s', name)
            process = mp.Process(target=handler.run)
            process.start()
            processes.append(process)

        for name, exchange in self._exchanges.items():
            LOGGER.info('Running exchange %s', name)

            if len(self._exchanges) > 1:
                process = mp.Process(target=exchange.run)
                process.start()
                processes.append(process)
            else:
                exchange.run()

        LOGGER.info('Joining all the processes')
        for process in processes:
            process.join()

    def archive(self, date):
        """Archive.
        """
        date = datetime.strptime(date, '%Y-%m-%d')

        LOGGER.info('Archiving the tables with date %s', date)

        processes = []

        for name, handler in self._handlers.items():
            LOGGER.info('Running handler %s', name)
            process = mp.Process(target=handler.run)
            process.start()
            processes.append(process)

        for exchange in self._exchanges.values():
            for name, instrument in exchange.instruments.items():
                for handler in exchange.handlers.values():
                    handler.rotate_table(
                        table=instrument,
                        last_datetime=date,
                        allow_fail=True)

        LOGGER.info('Closing the handlers')
        for handler in self._handlers.values():
            handler.prepare_close()

        LOGGER.info('Joining all the processes')
        for process in processes:
            process.join()

        LOGGER.info('Archived the tables with date %s', date)

    @staticmethod
    def create_exchange(
            exchange_name, subscription, handlers, is_debug, is_cold):
        """Create exchange.
        """
        try:
            from befh.exchange.websocket_exchange import WebsocketExchange
            exchange = WebsocketExchange(
                name=exchange_name,
                config=subscription,
                is_debug=is_debug,
                is_cold=is_cold)

            exchange.load(handlers=handlers)

        except ImportError as error:
            LOGGER.info(
                'Cannot load websocket exchange %s and fall into '
                'REST api exchange', exchange_name)
            from befh.exchange.rest_api_exchange import RestApiExchange
            exchange = RestApiExchange(
                name=exchange_name,
                config=subscription,
                is_debug=is_debug,
                is_cold=is_cold)

            exchange.load(handlers=handlers)

        return exchange

    @staticmethod
    def create_exchanges(
            exchanges_configuration, handlers, is_debug, is_cold):
        """Create exchanges.
        """
        exchanges = {}

        for exchange_name, subscription in exchanges_configuration.items():
            exchange = Runner.create_exchange(
                exchange_name=exchange_name,
                subscription=subscription,
                handlers=handlers,
                is_debug=is_debug,
                is_cold=is_cold)
            exchanges[exchange_name] = exchange

        return exchanges

    @staticmethod
    def create_handler(handler_name, handler_parameters, is_debug, is_cold):
        """Create handler.
        """
        LOGGER.info('Creating handler %s', handler_name)
        handler_name = handler_name.lower()

        if handler_name == "sql":
            from befh.handler import SqlHandler
            handler = SqlHandler(
                is_debug=is_debug,
                is_cold=is_cold,
                **handler_parameters)
        elif handler_name == "zmq":
            from befh.handler import ZmqHandler
            handler = ZmqHandler(
                is_debug=is_debug,
                is_cold=is_cold,
                **handler_parameters)
        else:
            raise NotImplementedError(
                'Handler %s is not implemented' % handler_name)

        handler.load(queue=mp.Queue())

        return handler

    @staticmethod
    def create_handlers(handlers_configuration, is_debug, is_cold):
        """Create handlers.
        """
        handlers = {}

        for handler_name, handler_para in handlers_configuration.items():
            handlers[handler_name] = Runner.create_handler(
                handler_name=handler_name,
                handler_parameters=handler_para,
                is_debug=is_debug,
                is_cold=is_cold)

        return handlers



import logging
from time import sleep

from .handler_operator import (
    HandlerOperator,
    HandlerCreateTableOperator,
    HandlerInsertOperator,
    HandlerRenameTableOperator,
    HandlerCloseOperator,
)

LOGGER = logging.getLogger(__name__)


class Handler:
    """Handler.
    """

    MAXIMUM_FAILURE_TOLERANCE = 2

    def __init__(self, is_debug, is_cold,
                 batch_frequency=1):
        """Constructor.
        """
        self._is_debug = is_debug
        self._is_cold = is_cold
        self._batch_frequency = batch_frequency
        self._is_running = False
        self._queue = None

    @property
    def is_rotate(self):
        """Is rotate.
        """
        return False

    @property
    def queue(self):
        """Queue.
        """
        return self._queue

    def load(self, queue):
        """Load.
        """
        LOGGER.info('Loading handler %s', self.__class__.__name__)
        self._queue = queue

    def prepare_create_table(self, table_name, fields, **kwargs):
        """Prepare create table.
        """
        self._queue.put(HandlerCreateTableOperator(
            table_name=table_name,
            fields=fields,
            **kwargs))

    def create_table(self, **kwargs):
        """Create table.
        """
        raise NotImplementedError(
            'Not implemented on handler %s' %
            self.__class__.__name__)

    def prepare_insert(self, table_name, fields, **kwargs):
        """Prepare insert.
        """
        self._queue.put(HandlerInsertOperator(
            table_name=table_name,
            fields=fields,
            **kwargs))

    def insert(self, **kwargs):
        """Insert.
        """
        raise NotImplementedError(
            'Not implemented on handler %s' %
            self.__class__.__name__)

    def prepare_rename_table(
            self, from_name, to_name, fields=None,
            keep_table=True, **kwargs):
        """Prepare rename table.
        """
        self._queue.put(HandlerRenameTableOperator(
            from_name=from_name,
            to_name=to_name,
            fields=fields,
            keep_table=keep_table,
            **kwargs))

    def rename_table(self, from_name, to_name, fields=None, keep_table=True):
        """Rename table.
        """
        raise NotImplementedError(
            'Not implemented on handler %s' %
            self.__class__.__name__)

    def update_order_book(self, exchange, symbol, bids, asks):
        """Update order book.
        """
        raise NotImplementedError(
            'Not implemented on exchange %s' %
            self.__class__.__name__)

    def update_trade(self, exchange, symbol, bids, asks):
        """Update trades.
        """
        raise NotImplementedError(
            'Not implemented on exchange %s' %
            self.__class__.__name__)

    def run(self):
        """Run.
        """
        LOGGER.info('Running %s', self.__class__.__name__)

        self._is_running = True

        while self._is_running:
            while not self._queue.empty():
                element = self._queue.get()
                assert isinstance(element, HandlerOperator), (
                    "Element type is not handler operator (%s)" % (
                        element.__class__.__name__))

                failure_count = 0

                while failure_count < self.MAXIMUM_FAILURE_TOLERANCE:
                    try:
                        element.execute(handler=self)
                        break
                    except Exception as exception:
                        failure_count += 1
                        if not self._should_rerun(element, exception):
                            # If the command should fail, the exception
                            # is raised within the method; otherwise
                            # it is the case that whether to rerun
                            break

                        LOGGER.warning(
                            'Element will be executed again due to the '
                            'failure with count %d', failure_count)

            sleep(self._batch_frequency)

        LOGGER.info('Completed running  %s', self.__class__.__name__)

    def prepare_close(self):
        """Close.
        """
        LOGGER.debug('Publishing close operator')
        self._queue.put(HandlerCloseOperator())

    def close(self):
        """Close.
        """
        LOGGER.debug('Publishing close operator')
        self._is_running = False

    def _should_rerun(self, element, exception):
        """Indicate whether the loop should rerun
        """
        if element.allow_fail:
            LOGGER.warn(
                'Execution failed on element %s (%s)',
                element,
                str(exception))
            return element.should_rerun
        else:
            raise



from time import sleep

import zmq

PORT = 9124


def main():
    """Main.
    """
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    print('Connect port %s' % PORT)
    socket.connect("tcp://localhost:%s" % PORT)

    while True:
        socket.send_json({'a': 1})
        sleep(5)
        socket.recv()


if __name__ == '__main__':
    main()



from datetime import datetime


class Field:
    """Field.
    """

    def __init__(self, name, value=None, is_key=False,
                 is_auto_increment=False):
        """Constructor.

        :param name: `str` of field name.
        :param value: `object` of value.
        :param is_key: `bool` indicating whether
            the field is a key.
        :param is_auto_increment: `bool` indicating the
            field is auto incremented in the handler.
        """
        self._name = name
        self._is_key = is_key
        self._value = value
        self._is_auto_increment = is_auto_increment

    @property
    def name(self):
        """Name.
        """
        return self._name

    @property
    def is_key(self):
        """Is key.
        """
        return self._is_key

    @property
    def is_auto_increment(self):
        """Auto increment.
        """
        return self._is_auto_increment

    @property
    def value(self):
        """Value.
        """
        return self._value

    @value.setter
    def value(self, value):
        """Set value.
        """
        self._value = value

    def __str__(self):
        """String.
        """
        return str(self._value)

    def __repr__(self):
        """Representation.
        """
        return "%s (%s)" % (
            self.__class__.__name__,
            self)

    def __eq__(self, other):
        """Equal.
        """
        if isinstance(other, Field):
            return self.value == other.value
        else:
            return self.value == other


class Table:
    """Table."""

    def __init__(self):
        """Constructor.
        """
        self._fields = None

    @classmethod
    def create_table(cls, handler):
        """Create table.
        """
        raise NotImplementedError(
            'Create table method is not implemented')

    @classmethod
    def insert(cls, **kwargs):
        """Insert.
        """
        raise NotImplementedError(
            'Insert table method is not implemented')

    def __str__(self):
        """String.
        """
        return self._fields


class IntIdField(Field):
    """Integer id field.
    """

    def __init__(self, name='id', value=1):
        """Constructor.
        """
        super().__init__(
            name=name, is_key=True, is_auto_increment=True,
            value=value)

    @property
    def field_type(self):
        """Field type.
        """
        return int

    def __gt__(self, other):
        """Greater than.
        """
        if isinstance(other, IntIdField):
            return self.value > other.value
        else:
            return self.value > other


class StringIdField(Field):
    """Integer id field.
    """

    def __init__(self, name='id', value=''):
        """Constructor.
        """
        super().__init__(
            name=name, is_key=True, value=value)

    @property
    def field_type(self):
        """Field type.
        """
        return str

    @property
    def field_length(self):
        """Field length.
        """
        return 64


class DateTimeField(Field):
    """Date time field.
    """

    @property
    def field_type(self):
        """Field type.
        """
        return datetime

    def __str__(self):
        """String.
        """
        return "'%s'" % self._value.strftime('%Y%m%d %H:%M:%S.%f')


class InstrumentNameField(Field):
    """Instrument name field.
    """

    @property
    def field_type(self):
        """Field type.
        """
        return str

    @property
    def field_length(self):
        """Field length.
        """
        return 20


class PriceField(Field):
    """Price field.
    """

    @property
    def field_type(self):
        """Field type.
        """
        return float

    @property
    def size(self):
        """Decimal size.
        """
        return 16

    @property
    def decimal(self):
        """Decimal place.
        """
        return 8


class QuantityField(Field):
    """Quantity field.
    """

    @property
    def field_type(self):
        """Field type.
        """
        return float

    @property
    def size(self):
        """Decimal size.
        """
        return 16

    @property
    def decimal(self):
        """Decimal place.
        """
        return 8
